/* ccgo: goboard.cc
 *
 * GoBoard widget definition
 * 
 * Copyright 2002 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <iostream>
#include "goboard.hh"
#define P(x, y) ((x) + (y) * z)

GoBoard::GoBoard(int sz) :
	z(sz),
	b(sz),
	m(sz)
{
	b.fill(NO_STONE);
	m.fill(NO_MARKING);
	add_events(GDK_POINTER_MOTION_MASK | GDK_ENTER_NOTIFY_MASK |
		   GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK |
		   GDK_BUTTON_RELEASE_MASK);
	t = TRACK_NONE;
	ton = false;
}

GoBoard::~GoBoard()
{
}

void GoBoard::put(const Place & p, BoardStates bs)
{
	b(p) = bs;
	draw(p.x, p.y);
}

void GoBoard::mark(const Place & p, Markings mk)
{
	m(p) = mk;
	draw(p.x, p.y);
}

void GoBoard::track(Trackings tk, BoardStates bs)
{
	if (ton) clear_track();
	t = tk;
	tf = bs;
}

void GoBoard::clear()
{
	b.fill(NO_STONE);
	m.fill(NO_MARKING);
	for (int x = 0; x < z; ++ x) for (int y = 0; y < z; ++ y)
		draw(x, y);
	ton = false;
}

BoardStates GoBoard::get(const Place & p)
{
	return b(p);
}

void GoBoard::realize_impl()
{
	DrawingArea::realize_impl();
	w = get_window();
	s = get_style();
}

gint GoBoard::expose_event_impl(GdkEventExpose * e)
{
	if (e->count > 0) return 1;
	for (int x = 0; x < z; ++ x) for (int y = 0; y < z; ++ y)
		draw(x, y);
	ton = false;
	return true;
}

gint GoBoard::enter_notify_event_impl(GdkEventCrossing * e)
{
	if (ton) clear_track();
	tx = int(e->x * z / width());
	ty = z - int(e->y * z / height()) - 1;
	pose(Place(tx, ty));
	if (t == TRACK_NONE) return true;
	draw_track();
	mbd = false;
	return true;
}

gint GoBoard::leave_notify_event_impl(GdkEventCrossing *)
{
	pose(Place(-1, -1));
	if (ton) clear_track();
	mbd = false;
	return true;
}

gint GoBoard::motion_notify_event_impl(GdkEventMotion * e)
{
	int x = int(e->x * z / width());
	int y = z - int(e->y * z / height()) - 1;
	pose(Place(x, y));
	if (t == TRACK_NONE) {
		if (ton) clear_track();
		return true;
	}
	if (x == tx && y == ty) {
		if (! ton) draw_track();
		return true;
	}
	mbd = false;
	draw(tx, ty);
	ton = false;
	tx = x;
	ty = y;
	draw_track();
	return true;
}

gint GoBoard::button_press_event_impl(GdkEventButton * e)
{
	if (t == TRACK_NONE) {
		if (ton) clear_track();
		return true;
	}
	if (ton) {
		mbd = true;
		draw_track();
	}
	return true;
}

gint GoBoard::button_release_event_impl(GdkEventButton * e)
{
	if (mbd && ton) {
		mbd = false;
		clear_track();
		hit.emit(Place(tx, ty));
	}
	return true;
}

void GoBoard::draw(int x, int y)
{
	int cx = (2 * x + 1) * width() / z / 2;
	int cy = (2 * (z - 1 - y) + 1) * height() / z / 2;
	int wx = width() / z - 1;
	int wy = height() / z - 1;
	switch (b(x, y)) {
	case W_STONE:
		w.draw_arc(s->get_white_gc(), true,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		w.draw_arc(s->get_black_gc(), false,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		break;
	case B_STONE:
		w.draw_arc(s->get_black_gc(), true,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		w.draw_arc(s->get_white_gc(), false,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		break;
	case NO_STONE:
		w.draw_arc(s->get_bg_gc(get_state()), true,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		w.draw_arc(s->get_bg_gc(get_state()), false,
			   (2 * cx - wx) / 2,
			   (2 * cy - wy) / 2,
			   wx,
			   wy,
			   0, 64 * 360);
		w.draw_line(s->get_black_gc(),
			    (x == 0 ? cx : (2 * cx - wx - 1) / 2),
			    cy,
			    (x == z - 1 ? cx : (2 * cx + wx + 1) / 2),
			    cy);
		w.draw_line(s->get_black_gc(),
			    cx,
			    (y == z - 1 ? cy : (2 * cy - wy - 1) / 2),
			    cx,
			    (y == 0 ? cy : (2 * cy + wy + 1) / 2));
		if ((x == 3 || x == 9 || x == 15) &&
		    (y == 3 || y == 9 || y == 15)) {
			w.draw_arc(s->get_black_gc(), true,
				   (2 * cx - wx / 3) / 2,
				   (2 * cy - wy / 3) / 2,
				   wx / 3,
				   wy / 3,
				   0, 64 * 360);
		}
	}
	switch(m(x, y)) {
	case NO_MARKING:
		return;
	case LAST_MOVE: // last stone
		w.draw_arc((b(x, y) == 1) ? s->get_white_gc()
			   : s->get_black_gc(), false,
			   (2 * cx - wx / 2) / 2,
			   (2 * cy - wy / 2) / 2,
			   wx / 2,
			   wy / 2,
			   0, 64 * 360);
		break;
	case KO_SPACE:
		w.draw_line(s->get_black_gc(),
			    cx + wx / 4, cy - wy / 4,
			    cx - wx / 4, cy + wy / 4);
		w.draw_line(s->get_black_gc(),
			    cx - wx / 4, cy - wy / 4,
			    cx + wx / 4, cy + wy / 4);
		break;
	case B_TERRI:
		w.draw_arc(s->get_black_gc(), true,
			   (4 * cx - wx) / 4,
			   (4 * cy - wy) / 4,
			   wx / 2,
			   wy / 2,
			   0, 64 * 360);
		break;
	case W_TERRI:
		w.draw_arc(s->get_white_gc(), true,
			   (4 * cx - wx) / 4,
			   (4 * cy - wy) / 4,
			   wx / 2,
			   wy / 2,
			   0, 64 * 360);
		break;
	default:
		;
	}
}


void GoBoard::draw_track()
{
	int cx = (2 * tx + 1) * width() / z / 2;
	int cy = (2 * (z - 1 - ty) + 1) * height() / z / 2;
	int wx = width() / z - 1;
	int wy = height() / z - 1;
	switch (t) {
	case TRACK_NONE:
		return;
	case TRACK_STONES:
		switch (b(tx, ty)) {
		case NO_STONE:
			return;
		case B_STONE:
			w.draw_arc(s->get_white_gc(), true,
				   (2 * cx - wx / 2) / 2,
				   (2 * cy - wy / 2) / 2,
				   wx / 2,
				   wy / 2,
				   0, 64 * 360);
			break;
		case W_STONE:
			w.draw_arc(s->get_black_gc(), true,
				   (2 * cx - wx / 2) / 2,
				   (2 * cy - wy / 2) / 2,
				   wx / 2,
				   wy / 2,
				   0, 64 * 360);
		}
		break;
	case TRACK_SPACE:
		switch (b(tx, ty)) {
		case B_STONE:
			return;
		case W_STONE:
			return;
		case NO_STONE:
			if (m(tx, ty) == KO_SPACE) return;
			switch (tf) {
			case NO_STONE:
				return;
			case B_STONE:
				w.draw_arc(s->get_black_gc(), true,
					   (2 * cx - wx) / 2,
					   (2 * cy - wy) / 2,
					   wx,
					   wy,
					   0, 64 * 360);
				if (mbd) w.draw_arc(s->get_white_gc(), false,
						    (2 * cx - wx) / 2,
						    (2 * cy - wy) / 2,
						    wx,
						    wy,
						    0, 64 * 360);
				break;
			case W_STONE:
				w.draw_arc(s->get_white_gc(), true,
					   (2 * cx - wx) / 2,
					   (2 * cy - wy) / 2,
					   wx,
					   wy,
					   0, 64 * 360);
				if (mbd) w.draw_arc(s->get_black_gc(), false,
						    (2 * cx - wx) / 2,
						    (2 * cy - wy) / 2,
						    wx,
						    wy,
						    0, 64 * 360);
			}
		}
		break;
	case TRACK_ALL:
		switch (b(tx, ty)) {
		case NO_STONE:
			w.draw_arc(s->get_black_gc(), true,
				   (2 * cx - wx / 2) / 2,
				   (2 * cy - wy / 2) / 2,
				   wx / 2,
				   wy / 2,
				   0, 64 * 360);
			break;
		case B_STONE:
			w.draw_arc(s->get_white_gc(), true,
				   (2 * cx - wx / 2) / 2,
				   (2 * cy - wy / 2) / 2,
				   wx / 2,
				   wy / 2,
				   0, 64 * 360);
			break;
		case W_STONE:
			w.draw_arc(s->get_black_gc(), true,
				   (2 * cx - wx / 2) / 2,
				   (2 * cy - wy / 2) / 2,
				   wx / 2,
				   wy / 2,
				   0, 64 * 360);
		}
	}
	ton = true;
}

void GoBoard::clear_track()
{
	draw(tx, ty);
	ton = false;
}
